import os
import urllib.parse
from datetime import datetime
from pathlib import Path
from typing import List

import click
import inquirer
import paramiko

from openapi_client import CliWorkspaceBackupCreateAPIPayload
from openapi_client.models import ResponseWorkspaceDetail, ResponseWorkspaceList
from vessl import vessl_api
from vessl.organization import _get_organization_name
from vessl.util.common import parse_time_to_ago
from vessl.util.config import VesslConfigLoader
from vessl.util.constant import (
    SSH_CONFIG_FORMAT,
    SSH_CONFIG_PATH, TEMP_DIR,
)
from vessl.util.exception import InvalidWorkspaceError, VesslException
from vessl.util.random import random_string
from vessl.util.zipper import Zipper
from vessl.volume import copy_volume_file


def read_workspace(workspace_id: int, **kwargs) -> ResponseWorkspaceDetail:
    """Read workspace

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.workspace_read_api(
        workspace_id=workspace_id, organization_name=_get_organization_name(**kwargs)
    )


def list_workspaces(
    cluster_id: int = None, statuses: List[str] = None, mine: bool = True, **kwargs
) -> List[ResponseWorkspaceList]:
    """List workspaces

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.workspace_list_api(
        organization_name=_get_organization_name(**kwargs),
        cluster=cluster_id,
        mine=mine,
        statuses=statuses,
    ).results


def backup_workspace() -> None:
    """Backup the home directory of the workspace

    Should only be called inside a workspace.
    """

    workspace_id = VesslConfigLoader().workspace
    if workspace_id is None:
        raise InvalidWorkspaceError("Can only be called within a workspace.")

    workspace = read_workspace(workspace_id)

    filename = f'workspace-backup-{datetime.utcnow().strftime("%Y%m%d%H%M%S")}-{random_string()}.zip'
    full_path = os.path.join(TEMP_DIR, filename)
    home_dir = str(Path.home())
    zipper = Zipper(full_path, "w")
    zipper.zipdir(home_dir, include_dotfiles=True)
    zipper.close()
    size = zipper.size()
    print(f"Backup file size: {size}")
    if size > 15 * 1024 * 1024 * 1024:
        zipper.remove()
        raise VesslException(f"{home_dir} is too large to backup.")

    print("Uploading the backup file...")
    copy_volume_file(
        source_volume_id=None,
        source_path=full_path,
        dest_volume_id=workspace.backup_volume_id,
        dest_path=filename,
    )
    zipper.remove()

    vessl_api.cli_workspace_backup_create_api(
        workspace_id=workspace_id,
        cli_workspace_backup_create_api_payload=CliWorkspaceBackupCreateAPIPayload(
            filename=filename,
        ),
    )


def restore_workspace(workspace_id: int = None) -> None:
    """Restore the home directory from the previous backup

    Should only be called inside a workspace.
    """

    if workspace_id is None:
        workspace_id = VesslConfigLoader().workspace
        if workspace_id is None:
            raise InvalidWorkspaceError("Can only be called within a workspace.")

        workspace_list = list_workspaces(mine=True)
        backup_workspaces = [w for w in workspace_list if w.last_backup is not None]
        if len(backup_workspaces) == 0:
            raise click.ClickException("Available workspace backup not found.")

        workspace = inquirer.prompt(
            [
                inquirer.List(
                    "question",
                    message="Select workspace",
                    choices=[
                        (
                            f"{w.name} (backup created {parse_time_to_ago(w.last_backup.created_dt)})",
                            w,
                        )
                        for w in backup_workspaces
                    ],
                )
            ],
            raise_keyboard_interrupt=True,
        ).get("question")
    else:
        workspace = read_workspace(workspace_id=workspace_id)
        if workspace.last_backup is None:
            raise click.ClickException("This workspace does not have any backup.")

    dest_path = os.path.join(TEMP_DIR, "workspace-backup.zip")
    print("Downloading the backup file...")
    copy_volume_file(
        source_volume_id=workspace.backup_volume_id,
        source_path=workspace.last_backup.filename,
        dest_volume_id=None,
        dest_path=dest_path,
    )

    zipper = Zipper(dest_path, "r")
    size = zipper.size()
    print(f"Backup file size: {size}")
    extract_path = str(Path.home())
    if size > 15 * 1024 * 1024 * 1024:
        extract_path = os.path.join(TEMP_DIR, "workspace-backup")
    print("Extracting...")
    zipper.extractall(extract_path)
    if size > 15 * 1024 * 1024 * 1024:
        print(f"Restored to {extract_path} since the size of the backup file is larger than 15Gi.")
    zipper.close()
    zipper.remove()


def connect_workspace_ssh(private_key_path: str, **kwargs) -> None:
    """Connect to a running workspace via SSH

    Keyword args:
        organization_name (str): override default organization
    """
    running_workspaces = list_workspaces(statuses=["running"], mine=True)
    if len(running_workspaces) == 0:
        raise click.ClickException("There is no running workspace.")

    if len(running_workspaces) == 1:
        workspace = running_workspaces[0]
    else:
        workspace = inquirer.prompt(
            [
                inquirer.List(
                    "question",
                    message="Select workspace",
                    choices=[
                        (f"{w.name} (created {parse_time_to_ago(w.created_dt)})", w)
                        for w in running_workspaces
                    ],
                )
            ],
            raise_keyboard_interrupt=True,
        ).get("question")

    ssh_endpoint = urllib.parse.urlparse(workspace.endpoints.ssh.endpoint)
    ssh_private_key_option = f" -i {private_key_path}" if private_key_path else ""
    os.system(
        f"ssh -p {ssh_endpoint.port}{ssh_private_key_option} vessl@{ssh_endpoint.hostname}"
    )


def update_vscode_remote_ssh(private_key_path: str) -> None:
    """Update .ssh/config file for VSCode Remote-SSH plugin"""
    running_workspaces = list_workspaces(statuses=["running"], mine=True)
    if len(running_workspaces) == 0:
        raise click.ClickException("There is no running workspace.")

    ssh_config = paramiko.SSHConfig()
    try:
        with open(SSH_CONFIG_PATH, "r") as fr:
            ssh_config.parse(fr)
    except FileNotFoundError:
        pass

    host_set = ssh_config.get_hostnames()
    for workspace in running_workspaces:
        host = f"{workspace.name}-{int(workspace.created_dt.timestamp())}"
        if host in host_set:
            continue

        ssh_endpoint = urllib.parse.urlparse(workspace.endpoints.ssh.endpoint)

        config_value = SSH_CONFIG_FORMAT.format(
            host=host,
            hostname=ssh_endpoint.hostname,
            port=ssh_endpoint.port,
        )
        if private_key_path:
            config_value += f"    IdentityFile {private_key_path}\n"

        with open(SSH_CONFIG_PATH, "a") as f:
            f.write(config_value)



from datetime import datetime

import click

from openapi_client import ResponseSimpleModelDetail
from vessl.cli._base import VesslGroup, vessl_argument
from vessl.cli._util import (
    Endpoint,
    format_url,
    generic_prompter,
    print_data,
    print_table,
    prompt_choices,
    truncate_datetime,
)
from vessl.cli.organization import organization_name_option
from vessl.model_repository import (
    create_model_repository,
    list_model_repositories,
    read_model_repository,
)


def model_repository_status(latest_model: ResponseSimpleModelDetail) -> str:
    if latest_model:
        return latest_model.status
    return "Empty repository"


def model_repository_update_dt(
    latest_model: ResponseSimpleModelDetail, default: datetime
) -> datetime:
    if latest_model:
        return truncate_datetime(latest_model.created_dt)
    return truncate_datetime(default)


def model_repository_name_callback(
    ctx: click.Context, param: click.Parameter, value: str
) -> str:
    if value:
        ctx.obj["repository"] = value
    return value


def model_repository_name_prompter(
    ctx: click.Context,
    param: click.Parameter,
    value: str,
) -> str:
    model_repositories = list_model_repositories()
    if len(model_repositories) == 0:
        raise click.UsageError(
            message="Create model repository with `vessl model-repository create`"
        )
    repository = prompt_choices(
        "Model repository", [x.name for x in model_repositories]
    )
    ctx.obj["repository"] = repository
    return repository


@click.command(name="model-repository", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
@vessl_argument(
    "name", type=click.STRING, required=True, prompter=model_repository_name_prompter
)
@organization_name_option
def read(name: str):
    model_repository = read_model_repository(repository_name=name)
    print_data(
        {
            "ID": model_repository.id,
            "Name": model_repository.name,
            "Description": model_repository.description,
            "Status": model_repository_status(
                model_repository.model_summary.latest_model
            ),
            "Organization": model_repository.organization.name,
            "Created": truncate_datetime(model_repository.created_dt),
            "Updated": truncate_datetime(model_repository.updated_dt),
        }
    )
    print(
        f"For more info: {format_url(Endpoint.model_repository.format(model_repository.organization.name, model_repository.name))}"
    )


@cli.vessl_command()
@organization_name_option
def list():
    model_repositories = list_model_repositories()
    print_table(
        model_repositories,
        ["Name", "Status", "Models", "Created", "Updated"],
        lambda x: [
            x.name,
            model_repository_status(x.model_summary.latest_model),
            x.model_summary.total,
            truncate_datetime(x.created_dt),
            model_repository_update_dt(x.model_summary.latest_model, x.updated_dt),
        ],
    )


@cli.vessl_command()
@vessl_argument(
    "name",
    type=click.STRING,
    required=True,
    prompter=generic_prompter("Model repository name"),
)
@click.option("-m", "--description", type=click.STRING)
@organization_name_option
def create(
    name: str,
    description: str,
):
    model_repository = create_model_repository(
        name=name,
        description=description,
    )
    print(
        f"Created '{model_repository.name}'.\n"
        f"For more info: {format_url(Endpoint.model_repository.format(model_repository.organization.name, model_repository.name))}"
    )



from typing import List

from openapi_client import ResponseFileMetadata
from openapi_client.models import (
    ModelCreateAPIPayload,
    ModelUpdateAPIPayload,
    ResponseModelDetail,
)
from vessl import vessl_api
from vessl.organization import _get_organization_name
from vessl.volume import copy_volume_file, delete_volume_file, list_volume_files


def read_model(
    repository_name: str, model_number: int, **kwargs
) -> ResponseModelDetail:
    """Read model

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_read_api(
        organization_name=_get_organization_name(**kwargs),
        repository_name=repository_name,
        number=model_number,
    )


def list_models(repository_name: str, **kwargs) -> List[ResponseModelDetail]:
    """List models

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_list_api(
        organization_name=_get_organization_name(**kwargs),
        repository_name=repository_name,
    ).results


def create_model(
    repository_name: str,
    repository_description: str = None,
    experiment_id: int = None,
    model_name: str = None,
    paths: List[str] = None,
    **kwargs,
) -> ResponseModelDetail:
    """Create model

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_create_api(
        organization_name=_get_organization_name(**kwargs),
        repository_name=repository_name,
        model_create_api_payload=ModelCreateAPIPayload(
            repository_description=repository_description,
            experiment_id=experiment_id,
            model_name=model_name,
            paths=paths,
        ),
    )


def update_model(
    repository_name: str, number: int, name: str, **kwargs
) -> ResponseModelDetail:
    """Update model

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_update_api(
        organization_name=_get_organization_name(**kwargs),
        repository_name=repository_name,
        number=number,
        model_update_api_payload=ModelUpdateAPIPayload(name=name),
    )


def delete_model(repository_name: str, number: int, **kwargs) -> object:
    """Delete model

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_delete_api(
        organization_name=_get_organization_name(**kwargs),
        repository_name=repository_name,
        version=number,
    )


def list_model_volume_files(
    repository_name: str,
    model_number: int,
    need_download_url: bool = False,
    path: str = "",
    recursive: bool = False,
    **kwargs,
) -> List[ResponseFileMetadata]:
    """List model files

    Keyword args:
        organization_name (str): override default organization
    """
    model = read_model(
        repository_name=repository_name, model_number=model_number, **kwargs
    )
    return list_volume_files(
        model.artifact_volume_id, need_download_url, path, recursive
    )


def upload_model_volume_file(
    repository_name: str,
    model_number: int,
    source_path: str,
    dest_path: str,
    **kwargs,
) -> None:
    """Upload file to model

    Keyword args:
        organization_name (str): override default organization
    """
    model = read_model(
        repository_name=repository_name, model_number=model_number, **kwargs
    )
    return copy_volume_file(
        source_volume_id=None,
        source_path=source_path,
        dest_volume_id=model.artifact_volume_id,
        dest_path=dest_path,
    )


def download_model_volume_file(
    repository_name: str,
    model_number: int,
    source_path: str,
    dest_path: str,
    **kwargs,
) -> None:
    """Download file to model

    Keyword args:
        organization_name (str): override default organization
    """
    model = read_model(
        repository_name=repository_name, model_number=model_number, **kwargs
    )
    return copy_volume_file(
        source_volume_id=model.artifact_volume_id,
        source_path=source_path,
        dest_volume_id=None,
        dest_path=dest_path,
    )


def delete_model_volume_file(
    repository_name: str,
    model_number: int,
    path: str,
    recursive: bool = False,
    **kwargs,
) -> List[ResponseFileMetadata]:
    """Delete model volume file

    Keyword args:
        organization_name (str): override default organization
    """
    model = read_model(
        repository_name=repository_name, model_number=model_number, **kwargs
    )
    return delete_volume_file(model.artifact_volume_id, path, recursive)



from typing import List

from openapi_client import ModelRepositoryUpdateAPIPayload
from openapi_client.models import (
    ModelRepositoryCreateAPIPayload,
    ResponseModelRepositoryDetail,
)
from vessl import vessl_api
from vessl.organization import _get_organization_name


def read_model_repository(
    repository_name: str, **kwargs
) -> ResponseModelRepositoryDetail:
    """Read model repository

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_repository_read_api(
        organization_name=_get_organization_name(**kwargs),
        repository_name=repository_name,
    )


def list_model_repositories(**kwargs) -> List[ResponseModelRepositoryDetail]:
    """List model repositories

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_repository_list_api(
        organization_name=_get_organization_name(**kwargs),
    ).results


def create_model_repository(
    name: str, description: str = None, **kwargs
) -> ResponseModelRepositoryDetail:
    """Create model repository

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_repository_create_api(
        organization_name=_get_organization_name(**kwargs),
        model_repository_create_api_payload=ModelRepositoryCreateAPIPayload(
            name=name,
            description=description,
        ),
    )


def update_model_repository(
    name: str, description: str, **kwargs
) -> ResponseModelRepositoryDetail:
    """Update model repository

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_update_api(
        organization_name=_get_organization_name(**kwargs),
        name=name,
        model_repository_update_api_payload=ModelRepositoryUpdateAPIPayload(
            description=description
        ),
    )


def delete_model_repository(name: str, **kwargs) -> object:
    """Delete model repository

    Keyword args:
        organization_name (str): override default organization
    """
    return vessl_api.model_repository_delete_api(
        organization_name=_get_organization_name(**kwargs),
        name=name,
    )
